"""Template constants for pipeline builder."""

# Import all template modules to make them accessible from the templates package
from .application_runner import (
    APPLICATION_RUNNER_TEMPLATE,
)
from .b_unit import B_UNIT_TEMPLATE
from .b_unit_creator_and_runner import (
    B_UNIT_CREATOR_AND_RUNNER_TEMPLATE,
)
from .conftest import CONFTEST_TEMPLATE
from .e2e_test import E2E_TEST_TEMPLATE
from .pipeline_orchestrator import (
    EMPTY_FUNCTION_BODY as PIPELINE_EMPTY_FUNCTION_BODY,
)
from .pipeline_orchestrator import (
    ORCHESTRATE_THIN_SLICE_CALL,
    PIPELINE_ORCHESTRATOR_FUNCTION_START,
    PIPELINE_ORCHESTRATOR_IMPORT_TEMPLATE,
)
from .pipeline_runner import (
    PIPELINE_RUNNER_TEMPLATE,
)
from .pipelines_runner import (
    EMPTY_FUNCTION_BODY as PIPELINES_EMPTY_FUNCTION_BODY,
)
from .pipelines_runner import (
    PIPELINE_IMPORT_TEMPLATE,
    PIPELINE_RUNNER_CALL,
    PIPELINES_RUNNER_FUNCTION_START,
    PIPELINES_RUNNER_TEMPLATE_HEADER,
)
from .stage_orchestrator import (
    B_UNIT_CREATOR_IMPORT as STAGE_B_UNIT_CREATOR_IMPORT,
)
from .stage_orchestrator import (
    B_UNIT_IMPORT_TEMPLATE,
)
from .stage_orchestrator import (
    CREATE_AND_RUN_B_UNIT_CALL as STAGE_CREATE_AND_RUN_B_UNIT_CALL,
)
from .stage_orchestrator import (
    EMPTY_FUNCTION_BODY as STAGE_EMPTY_FUNCTION_BODY,
)
from .stage_orchestrator import (
    ORCHESTRATE_SUB_STAGE_CALL,
    STAGE_ORCHESTRATOR_FUNCTION_START,
    STAGE_ORCHESTRATOR_HEADER,
    SUB_STAGE_IMPORT_TEMPLATE,
)
from .sub_stage_orchestrator import (
    B_UNIT_CREATOR_IMPORT as SUB_STAGE_B_UNIT_CREATOR_IMPORT,
)
from .sub_stage_orchestrator import (
    CREATE_AND_RUN_B_UNIT_CALL as SUB_STAGE_CREATE_AND_RUN_B_UNIT_CALL,
)
from .sub_stage_orchestrator import (
    EMPTY_FUNCTION_BODY as SUB_STAGE_EMPTY_FUNCTION_BODY,
)
from .sub_stage_orchestrator import (
    SUB_STAGE_B_UNIT_IMPORT_TEMPLATE,
    SUB_STAGE_ORCHESTRATOR_FUNCTION_START,
    SUB_STAGE_ORCHESTRATOR_HEADER,
)
from .thin_slice_orchestrator import (
    EMPTY_FUNCTION_BODY as THIN_SLICE_EMPTY_FUNCTION_BODY,
)
from .thin_slice_orchestrator import (
    ORCHESTRATE_STAGE_CALL,
    THIN_SLICE_ORCHESTRATOR_FUNCTION_START,
    THIN_SLICE_ORCHESTRATOR_IMPORT_TEMPLATE,
)



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def add_new_attribute_to_dictionary(
    new_ea_attributes_dictionary: dict,
    name_object_nf_uuid: str,
    attribute_value: str,
    attribute_name: str,
    attribute_type_nf_uuid: str,
) -> str:
    attribute_nf_uuid = (
        create_new_uuid()
    )

    attribute_ea_guid = (
        create_ea_guid_from_nf_uuid(
            nf_uuid=attribute_nf_uuid,
        )
    )

    new_ea_attribute_dictionary = {
        NfColumnTypes.NF_UUIDS.column_name: attribute_nf_uuid,
        NfEaComColumnTypes.ELEMENT_COMPONENTS_DEFAULT.column_name: attribute_value,
        NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name: name_object_nf_uuid,
        NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name: attribute_type_nf_uuid,
        NfEaComColumnTypes.ELEMENT_COMPONENTS_UML_VISIBILITY_KIND.column_name: "Public",
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: attribute_name,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: attribute_ea_guid,
    }

    new_ea_attributes_dictionary[
        attribute_nf_uuid
    ] = new_ea_attribute_dictionary

    return attribute_nf_uuid



from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.substages.operations.a_load.content_operations.digitalisation_levels.digitalisation_level_stereotype_replacer import (
    replace_digitalisation_level_stereotype,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)


def shift_convention_universe_to_semantically_grounded(
    content_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
) -> NfEaComUniverses:
    log_message(
        message="Shift Universe to Semantically Grounded Digitalisation Level - started",
    )

    output_universe.nf_ea_com_registry.dictionary_of_collections = (
        content_universe.nf_ea_com_registry.dictionary_of_collections.copy()
    )

    replace_digitalisation_level_stereotype(
        nf_ea_com_universe=output_universe,
        digitalisation_level_stereotype=DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_8_CLASS_STEREOTYPE,
    )

    log_message(
        message="Shift Universe to Semantically Grounded Digitalisation Level - finished",
    )

    return output_universe



"""Template for e2e test."""

E2E_TEST_TEMPLATE = """import pytest
from {domain_name}.b_source.app_runners.{domain_name}_b_clearer_pipeline_b_application_runner import (
    run_{domain_name}_b_clearer_pipeline_b_application,
)


def test_{domain_name}_b_clearer_pipeline_b_application(e2e_test_setup, e2e_test_teardown):
    # Run the pipeline
    run_{domain_name}_b_clearer_pipeline_b_application()

    # Add assertions here
    assert True
"""



import json
import os
import shutil
import tempfile
from pathlib import Path

import pytest
from bclearer_core.pipeline_builder.generator import (
    generate_pipeline,
)
from bclearer_core.pipeline_builder.schema import (
    get_sample_config,
)


class TestPipelineGenerationFlow:
    """Test class for end-to-end pipeline generation flow."""

    @pytest.fixture(autouse=True)
    def setup_method(
        self,
        pipeline_output_folder_absolute_path,
    ):
        """Setup method that runs before each test."""
        # Create a temporary directory for output
        self.temp_output_dir = os.path.join(
            pipeline_output_folder_absolute_path,
            "temp_flow_test",
        )
        os.makedirs(
            self.temp_output_dir,
            exist_ok=True,
        )

        # Clean up any existing test pipelines from previous runs
        self.test_domain_path = os.path.join(
            self.temp_output_dir,
            "example_domain_pipelines",
        )

        if os.path.exists(
            self.test_domain_path
        ):
            shutil.rmtree(
                self.test_domain_path
            )

        # Also clean up custom_domain directory for the custom test
        self.custom_domain_path = os.path.join(
            self.temp_output_dir,
            "custom_domain_pipelines",
        )

        if os.path.exists(
            self.custom_domain_path
        ):
            shutil.rmtree(
                self.custom_domain_path
            )

    def teardown_method(self):
        """Cleanup after tests."""
        if os.path.exists(
            self.temp_output_dir
        ):
            shutil.rmtree(
                self.temp_output_dir
            )

    def test_full_pipeline_generation_flow(
        self,
        template_pipeline_folder_absolute_path,
    ):
        """
        Test the complete flow from sample config to pipeline generation,
        verifying the structure and runability of the generated pipeline.
        """
        # Get sample config
        sample_config = (
            get_sample_config()
        )

        # Generate the pipeline
        pipeline_path = (
            generate_pipeline(
                sample_config,
                self.temp_output_dir,
            )
        )

        # Verify pipeline structure
        domain_path = os.path.join(
            self.temp_output_dir,
            "example_domain_pipelines",
        )
        assert os.path.exists(
            domain_path
        )

        # Check key top-level directories
        assert os.path.exists(
            os.path.join(
                domain_path, "b_source"
            )
        )
        assert os.path.exists(
            os.path.join(
                domain_path, "resources"
            )
        )
        assert os.path.exists(
            os.path.join(
                domain_path, "tests"
            )
        )

        # Check app_runners directory structure
        app_runners_path = os.path.join(
            domain_path,
            "b_source",
            "app_runners",
        )
        assert os.path.exists(
            app_runners_path
        )
        assert os.path.exists(
            os.path.join(
                app_runners_path,
                "runners",
            )
        )
        assert os.path.exists(
            os.path.join(
                app_runners_path,
                f"{sample_config['domain_name']}_b_clearer_pipeline_b_application_runner.py",
            )
        )

        # Check common directory structure
        common_path = os.path.join(
            domain_path,
            "b_source",
            "common",
        )
        assert os.path.exists(
            common_path
        )
        assert os.path.exists(
            os.path.join(
                common_path,
                "operations",
                "b_units",
            )
        )

        # Verify the pipeline directory structure
        pipeline_name = sample_config[
            "pipelines"
        ][0]["name"]
        pipeline_path = os.path.join(
            domain_path,
            "b_source",
            pipeline_name,
        )
        assert os.path.exists(
            pipeline_path
        )

        # Check pipeline subdirectories
        assert os.path.exists(
            os.path.join(
                pipeline_path, "objects"
            )
        )
        assert os.path.exists(
            os.path.join(
                pipeline_path,
                "operations",
            )
        )
        assert os.path.exists(
            os.path.join(
                pipeline_path,
                "orchestrators",
            )
        )

        # Check orchestrator subdirectories
        orchestrators_path = (
            os.path.join(
                pipeline_path,
                "orchestrators",
            )
        )
        assert os.path.exists(
            os.path.join(
                orchestrators_path,
                "pipeline",
            )
        )
        assert os.path.exists(
            os.path.join(
                orchestrators_path,
                "stages",
            )
        )
        assert os.path.exists(
            os.path.join(
                orchestrators_path,
                "thin_slices",
            )
        )
        assert os.path.exists(
            os.path.join(
                orchestrators_path,
                "sub_stages",
            )
        )

        # Check the specific orchestrator files exist
        assert os.path.exists(
            os.path.join(
                orchestrators_path,
                "pipeline",
                f"{pipeline_name}_orchestrator.py",
            )
        )

        # Verify thin slice generation
        thin_slice_name = sample_config[
            "pipelines"
        ][0]["thin_slices"][0]["name"]
        assert os.path.exists(
            os.path.join(
                orchestrators_path,
                "thin_slices",
                f"{thin_slice_name}_orchestrator.py",
            )
        )

        # Verify stage generation
        for stage in sample_config[
            "pipelines"
        ][0]["thin_slices"][0][
            "stages"
        ]:
            stage_name = stage["name"]
            stage_orchestrator_name = f"{pipeline_name}_{stage_name}_orchestrator.py"
            assert os.path.exists(
                os.path.join(
                    orchestrators_path,
                    "stages",
                    stage_orchestrator_name,
                )
            )

            # If the stage has b_units, verify their generation
            if stage["b_units"]:
                stage_b_units_path = os.path.join(
                    pipeline_path,
                    "objects",
                    "b_units",
                    f"{pipeline_name}_{stage_name}",
                )
                assert os.path.exists(
                    stage_b_units_path
                )

                for b_unit in stage[
                    "b_units"
                ]:
                    b_unit_file_name = f"{b_unit.lower()}_b_units.py"
                    assert os.path.exists(
                        os.path.join(
                            stage_b_units_path,
                            b_unit_file_name,
                        )
                    )

            # If the stage has sub_stages, verify their generation
            for sub_stage in stage.get(
                "sub_stages", []
            ):
                sub_stage_name = (
                    sub_stage["name"]
                )
                sub_stage_path = os.path.join(
                    orchestrators_path,
                    "sub_stages",
                    f"{pipeline_name}_{stage_name}_{sub_stage_name}",
                )
                assert os.path.exists(
                    sub_stage_path
                )

                # Check sub_stage orchestrator
                sub_stage_orchestrator_name = f"{pipeline_name}_{stage_name}_{sub_stage_name}_orchestrator.py"
                assert os.path.exists(
                    os.path.join(
                        sub_stage_path,
                        sub_stage_orchestrator_name,
                    )
                )

                # Check sub_stage b_units
                if sub_stage["b_units"]:
                    sub_stage_b_units_path = os.path.join(
                        pipeline_path,
                        "objects",
                        "b_units",
                        f"{pipeline_name}_{stage_name}",
                        f"{pipeline_name}_{stage_name}_{sub_stage_name}",
                    )
                    assert os.path.exists(
                        sub_stage_b_units_path
                    )

                    for (
                        b_unit
                    ) in sub_stage[
                        "b_units"
                    ]:
                        b_unit_file_name = f"{b_unit.lower()}_b_units.py"
                        assert os.path.exists(
                            os.path.join(
                                sub_stage_b_units_path,
                                b_unit_file_name,
                            )
                        )

        # Verify test directory structure
        tests_path = os.path.join(
            domain_path, "tests"
        )
        assert os.path.exists(
            os.path.join(
                tests_path, "common"
            )
        )
        assert os.path.exists(
            os.path.join(
                tests_path, "universal"
            )
        )

        # Check E2E test setup
        e2e_path = os.path.join(
            tests_path,
            "universal",
            "e2e",
        )
        assert os.path.exists(
            os.path.join(
                e2e_path, "conftest.py"
            )
        )
        assert os.path.exists(
            os.path.join(
                e2e_path,
                f"test_{sample_config['domain_name']}_b_clearer_pipeline_b_application_runner.py",
            )
        )

    def test_custom_pipeline_generation(
        self,
        template_pipeline_folder_absolute_path,
    ):
        """Test generating a pipeline with a custom configuration."""
        # Create a custom pipeline config
        custom_config = {
            "domain_name": "custom_domain",
            "pipelines": [
                {
                    "name": "data_processing",
                    "thin_slices": [
                        {
                            "name": "data_extraction",
                            "stages": [
                                {
                                    "name": "1c_collect",
                                    "sub_stages": [
                                        {
                                            "name": "api_extraction",
                                            "b_units": [
                                                "api_reader",
                                                "api_parser",
                                            ],
                                        },
                                        {
                                            "name": "file_extraction",
                                            "b_units": [
                                                "csv_reader",
                                                "json_reader",
                                            ],
                                        },
                                    ],
                                    "b_units": [
                                        "extraction_coordinator"
                                    ],
                                },
                                {
                                    "name": "2l_load",
                                    "sub_stages": [],
                                    "b_units": [
                                        "data_loader",
                                        "data_validator",
                                    ],
                                },
                            ],
                        }
                    ],
                }
            ],
        }

        # Generate the pipeline
        pipeline_path = (
            generate_pipeline(
                custom_config,
                self.temp_output_dir,
            )
        )

        # Verify pipeline structure
        domain_path = os.path.join(
            self.temp_output_dir,
            "custom_domain_pipelines",
        )
        assert os.path.exists(
            domain_path
        )

        # Check data_processing pipeline exists
        data_processing_path = (
            os.path.join(
                domain_path,
                "b_source",
                "data_processing",
            )
        )
        assert os.path.exists(
            data_processing_path
        )

        # Check thin slice exists
        thin_slice_path = os.path.join(
            data_processing_path,
            "orchestrators",
            "thin_slices",
            "data_extraction_orchestrator.py",
        )
        assert os.path.exists(
            thin_slice_path
        )

        # Check collect stage with sub-stages
        collect_path = os.path.join(
            data_processing_path,
            "orchestrators",
            "stages",
            "data_processing_1c_collect_orchestrator.py",
        )
        assert os.path.exists(
            collect_path
        )

        # Check api_extraction sub-stage
        api_extraction_path = os.path.join(
            data_processing_path,
            "orchestrators",
            "sub_stages",
            "data_processing_1c_collect_api_extraction",
            "data_processing_1c_collect_api_extraction_orchestrator.py",
        )
        assert os.path.exists(
            api_extraction_path
        )

        # Check file_extraction sub-stage
        file_extraction_path = os.path.join(
            data_processing_path,
            "orchestrators",
            "sub_stages",
            "data_processing_1c_collect_file_extraction",
            "data_processing_1c_collect_file_extraction_orchestrator.py",
        )
        assert os.path.exists(
            file_extraction_path
        )

        # Check b_units for collect stage
        collect_b_units_path = os.path.join(
            data_processing_path,
            "objects",
            "b_units",
            "data_processing_1c_collect",
        )
        assert os.path.exists(
            collect_b_units_path
        )
        assert os.path.exists(
            os.path.join(
                collect_b_units_path,
                "extraction_coordinator_b_units.py",
            )
        )

        # Check b_units for api_extraction sub-stage
        api_extraction_b_units_path = os.path.join(
            data_processing_path,
            "objects",
            "b_units",
            "data_processing_1c_collect",
            "data_processing_1c_collect_api_extraction",
        )
        assert os.path.exists(
            api_extraction_b_units_path
        )
        assert os.path.exists(
            os.path.join(
                api_extraction_b_units_path,
                "api_reader_b_units.py",
            )
        )
        assert os.path.exists(
            os.path.join(
                api_extraction_b_units_path,
                "api_parser_b_units.py",
            )
        )

        # Check file contents for a b_unit
        with open(
            os.path.join(
                api_extraction_b_units_path,
                "api_reader_b_units.py",
            ),
            "r",
        ) as f:
            content = f.read()
            assert (
                "class ApiReaderBUnits:"
                in content
            )
            assert (
                "def run(self) -> None:"
                in content
            )
            assert (
                "def b_unit_process_function"
                in content
            )



from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)
from bclearer_core.substages.operations.common.connector_adder import (
    add_new_connector_to_dictionary,
)
from bclearer_core.substages.operations.common.instances_nf_uuids_getter import (
    get_instances_nf_uuids_of_matched_type,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)


def add_dependency_to_instances_of_type(
    content_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    matched_target_object: MatchedEaObjects,
    matched_source_objects_type: MatchedEaObjects,
):
    output_universe.nf_ea_com_registry.dictionary_of_collections = (
        content_universe.nf_ea_com_registry.dictionary_of_collections.copy()
    )

    matched_target_object_type_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=content_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=matched_target_object.ea_guid,
    )

    classifiers_instances_of_matched_source_objects_type = get_instances_nf_uuids_of_matched_type(
        nf_ea_com_universe=content_universe,
        matched_type=matched_source_objects_type,
    )

    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    for (
        instance_of_source_objects_type
    ) in classifiers_instances_of_matched_source_objects_type:
        add_new_connector_to_dictionary(
            new_connector_dictionary=new_ea_objects_dictionary[
                NfEaComCollectionTypes.EA_CONNECTORS
            ],
            place_1_nf_uuid=instance_of_source_objects_type,
            place_2_nf_uuid=matched_target_object_type_nf_uuid,
            connector_type=EaConnectorTypes.DEPENDENCY,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=output_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )



from bclearer_core.configurations.b_configurations.b_configurations import (
    BConfigurations,
)
from bclearer_core.configurations.datastructure.logging_inspection_level_b_enums import (
    LoggingInspectionLevelBEnums,
)
from bclearer_orchestration_services.log_environment_utility_service.common_knowledge.environment_log_level_types import (
    EnvironmentLogLevelTypes,
)
from bclearer_orchestration_services.logging_service.configuration.process_logging_levels import (
    ProcessLoggingLevels,
)


class BLoggingConfigurations(
    BConfigurations
):
    # TODO: should this be None instead?
    #  how could we switch between log_process_message() and log_message() in @run_and_log_function()?
    LOG_PROCESS_MESSAGES_FLAG = True

    console_process_logging_level = (
        ProcessLoggingLevels.PROCESS_LEVEL_10
    )

    file_process_logging_level = (
        ProcessLoggingLevels.PROCESS_LEVEL_2
    )

    # TODO: default level is WARNING - to agree
    LOGGING_INSPECTION_REPORTING_LEVEL_ENUM = (
        LoggingInspectionLevelBEnums.WARNING
    )

    # TODO: default level is ERROR - to agree
    LOGGING_INSPECTION_SUMMARISING_LEVEL_ENUM = (
        LoggingInspectionLevelBEnums.ERROR
    )

    ENVIRONMENT_LOG_LEVEL_TYPE = (
        EnvironmentLogLevelTypes.FILTERED
    )



from bclearer_core.common_knowledge.adjustment_operation_types import (
    AdjustmentOperationTypes,
)
from bclearer_core.common_knowledge.attribute_to_associations_operation_subtypes import (
    AttributeToAssociationOperationSubtypes,
)
from bclearer_core.configurations.adjustment_operation_configurations import (
    AdjustmentOperationConfigurations,
)
from bclearer_core.configurations.load_hdf5_model_configurations import (
    LoadHdf5ModelConfigurations,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_association_direction_types import (
    EaAssociationDirectionTypes,
)


class AttributeToAssociationAdjustmentOperationConfigurations(
    AdjustmentOperationConfigurations,
):
    def __init__(
        self,
        adjustment_operation_type: AdjustmentOperationTypes,
        adjustment_universe_load_hdf5_model_configuration: LoadHdf5ModelConfigurations,
        output_universe_short_name: str,
        direction: EaAssociationDirectionTypes,
        package_name: str,
        # attribute_to_association_operation_subtype: AttributeToAssociationOperationSubtypes,
    ):
        super().__init__(
            adjustment_operation_type=adjustment_operation_type,
            adjustment_universe_load_hdf5_model_configuration=adjustment_universe_load_hdf5_model_configuration,
            output_universe_short_name=output_universe_short_name,
        )

        self.direction = direction

        self.package_name = package_name

    #        self.attribute_to_association_operation_subtype = attribute_to_association_operation_subtype

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from b_clearer_base.b_source.universes.b_clearer_domain_universes import (
    BClearerDomainUniverses,
)


class BClearerRunUniverses(
    BClearerDomainUniverses
):
    def __init__(self):
        pass



class AdjustmentOperationsSubstageConfigurations:
    def __init__(
        self,
        operation_configurations: set,
    ):
        self.operation_configurations = (
            operation_configurations
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from enum import Enum, unique


@unique
class MatchedEaObjects(Enum):
    def __object_name(self) -> str:
        raise NotImplementedError

    def __ea_guid(self) -> str:
        raise NotImplementedError

    object_name = property(
        fget=__object_name,
    )

    ea_guid = property(fget=__ea_guid)



from bclearer_core.substages.operations.b_evolve.common.universe_aligner import (
    align_universes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)

ALIGNED_UNIVERSE_SHORT_NAME_SUFFIX = (
    "_aligned"
)


class UniversesMergeRegisters:
    def __init__(
        self,
        universe_1: NfEaComUniverses,
        universe_2: NfEaComUniverses,
        context: str,
    ):
        (
            primary_universe,
            aligned_universe,
        ) = UniversesMergeRegisters.__classify_universes(
            universe_1=universe_1,
            universe_2=universe_2,
        )

        self.primary_universe = (
            primary_universe
        )

        self.aligned_universe = (
            aligned_universe
        )

        self.aligned_to_primary_universe_nf_uuids_map = (
            dict()
        )

        align_universes(
            universe_merge_register=self,
            context=context,
        )

    @staticmethod
    def __classify_universes(
        universe_1: NfEaComUniverses,
        universe_2: NfEaComUniverses,
    ):
        # TODO This may be calculated on the size of universes if needed
        primary_universe = universe_1

        aligned_universe_short_name = (
            universe_2.ea_repository.short_name
            + ALIGNED_UNIVERSE_SHORT_NAME_SUFFIX
        )

        aligned_universe = universe_2.copy(
            short_name=aligned_universe_short_name,
        )

        return (
            primary_universe,
            aligned_universe,
        )



from bclearer_core.configurations.uml_name_to_named_object_configuration_objects import (
    UmlNameToNamedObjectConfigurationObjects,
)
from bclearer_core.constants.standard_constants import (
    DEFAULT_NULL_VALUE,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.common.all_classifiers_of_a_package_and_its_subpackages_getter import (
    get_all_package_and_subpackages_classifiers,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.common.has_classifier_ea_guid_attribute_type_name_and_value_checker import (
    has_classifier_ea_guid_attribute_type_name_and_value,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import DataFrame, concat


def shift_convention_uml_names_to_named_objects(
    content_universe: NfEaComUniverses,
    list_of_configuration_objects: list,
    output_universe: NfEaComUniverses,
) -> NfEaComUniverses:
    log_message(
        message="CONVENTION SHIFT OPERATION: Extract UML names to attributes - started",
    )

    output_universe.nf_ea_com_registry.dictionary_of_collections = (
        content_universe.nf_ea_com_registry.dictionary_of_collections.copy()
    )

    for (
        configuration_object
    ) in list_of_configuration_objects:
        __shift_object_uml_name_to_named_object(
            configuration_object=configuration_object,
            output_universe=output_universe,
        )

    log_message(
        message="CONVENTION SHIFT OPERATION: Extract UML names to attributes - finished",
    )

    return output_universe


def __shift_object_uml_name_to_named_object(
    configuration_object: UmlNameToNamedObjectConfigurationObjects,
    output_universe: NfEaComUniverses,
):
    package_guid = (
        configuration_object.matched_package.ea_guid
    )

    naming_space_guid = (
        configuration_object.matched_naming_space.ea_guid
    )

    # Only classes. Notes and boundaries (for example) are left out
    table_of_all_classifiers_of_a_package_and_its_subpackages = get_all_package_and_subpackages_classifiers(
        content_universe=output_universe,
        package_ea_guid=package_guid,
    )

    for (
        index,
        domain_object_row,
    ) in (
        table_of_all_classifiers_of_a_package_and_its_subpackages.iterrows()
    ):
        name_string = domain_object_row[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
        ]

        naming_space = (
            configuration_object.matched_naming_space.object_name
        )

        classifier_ea_guid = domain_object_row[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
        ]

        if not has_classifier_ea_guid_attribute_type_name_and_value(
            content_universe=output_universe,
            classifier_ea_guid=classifier_ea_guid,
            attribute_type_name=naming_space,
            attribute_value=name_string,
        ):
            __add_attribute_value_with_naming_space_type_to_output_universe(
                output_universe=output_universe,
                classifier_ea_guid=classifier_ea_guid,
                attribute_value=name_string,
                naming_space_ea_guid=naming_space_guid,
                naming_space=naming_space,
            )


def __add_attribute_value_with_naming_space_type_to_output_universe(
    output_universe: NfEaComUniverses,
    classifier_ea_guid: str,
    attribute_value: str,
    naming_space_ea_guid: str,
    naming_space: str,
):
    output_universe_ea_attributes = output_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.EA_ATTRIBUTES
    ]

    output_universe_ea_classifiers = output_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.EA_CLASSIFIERS
    ]

    ea_guid_column_name = (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
    )

    nf_uuids_column_name = (
        NfColumnTypes.NF_UUIDS.column_name
    )

    classifier_nf_uuid = (
        output_universe_ea_classifiers[
            output_universe_ea_classifiers[
                ea_guid_column_name
            ]
            == classifier_ea_guid
        ][nf_uuids_column_name]
        .to_string(index=False)
        .strip()
    )

    naming_space_nf_uuid = (
        output_universe_ea_classifiers[
            output_universe_ea_classifiers[
                ea_guid_column_name
            ]
            == naming_space_ea_guid
        ][nf_uuids_column_name]
        .to_string(index=False)
        .strip()
    )

    new_attribute_nf_uuid = (
        create_new_uuid()
    )

    new_attribute_ea_guid = create_ea_guid_from_nf_uuid(
        nf_uuid=new_attribute_nf_uuid,
    )

    # Create a new DataFrame row from the dictionary
    new_attribute_row = DataFrame(
        [
            {
                NfColumnTypes.NF_UUIDS.column_name: new_attribute_nf_uuid,
                NfEaComColumnTypes.ATTRIBUTES_LOWER_BOUNDS.column_name: "1",
                NfEaComColumnTypes.ATTRIBUTES_UPPER_BOUNDS.column_name: "1",
                NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name: classifier_nf_uuid,
                NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name: naming_space_nf_uuid,
                NfEaComColumnTypes.ELEMENT_COMPONENTS_UML_VISIBILITY_KIND.column_name: "Public",
                NfEaComColumnTypes.ELEMENT_COMPONENTS_TYPE.column_name: naming_space,
                NfEaComColumnTypes.ELEMENT_COMPONENTS_DEFAULT.column_name: DEFAULT_NULL_VALUE,
                NfEaComColumnTypes.STEREOTYPEABLE_OBJECTS_EA_OBJECT_STEREOTYPES.column_name: [],
                NfEaComColumnTypes.REPOSITORIED_OBJECTS_EA_REPOSITORY.column_name: DEFAULT_NULL_VALUE,
                NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: attribute_value,
                NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NOTES.column_name: DEFAULT_NULL_VALUE,
                NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: new_attribute_ea_guid,
            }
        ]
    )

    # Use pd.concat to append the new row to the existing DataFrame
    output_universe_ea_attributes = concat(
        [
            output_universe_ea_attributes,
            new_attribute_row,
        ],
        ignore_index=True,
    )

    output_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.EA_ATTRIBUTES
    ] = output_universe_ea_attributes



import json
import os
import shutil
from pathlib import Path

import pytest
from bclearer_core.pipeline_builder.generator import (
    generate_pipeline,
)
from bclearer_core.pipeline_builder.schema import (
    validate_pipeline_config,
)


class TestPipelineBuilder:
    """Test class for the pipeline builder functionality."""

    @pytest.fixture(autouse=True)
    def setup_method(
        self,
        pipeline_output_folder_absolute_path,
        pipeline_builder_configuration_file_path,
    ):
        """Setup method that runs before each test."""
        # Load the configuration to get domain_name
        with open(
            pipeline_builder_configuration_file_path,
            "r",
        ) as f:
            config_dict = json.load(f)

        # Get the domain name from the config
        domain_name = config_dict[
            "domain_name"
        ]

        # Clean up any existing test pipelines from previous runs
        self.domain_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "bclearer_pipelines",
            domain_name,
        )

        if os.path.exists(
            self.domain_path
        ):
            shutil.rmtree(
                self.domain_path
            )

    def test_validate_pipeline_config_valid_config(
        self,
        pipeline_builder_configuration_file_path,
    ):
        """Test that a valid configuration file is correctly validated."""
        # Load the test configuration
        with open(
            pipeline_builder_configuration_file_path,
            "r",
        ) as f:
            config_dict = json.load(f)

        # Validate the configuration
        config = (
            validate_pipeline_config(
                config_dict
            )
        )

        # Assert basic structure is correct
        assert (
            config.domain_name
            == config_dict[
                "domain_name"
            ]
        )
        assert len(
            config.pipelines
        ) == len(
            config_dict["pipelines"]
        )
        assert (
            config.pipelines[0].name
            == config_dict["pipelines"][
                0
            ]["name"]
        )
        assert (
            len(
                config.pipelines[
                    0
                ].thin_slices
            )
            == 1
        )
        assert (
            len(
                config.pipelines[0]
                .thin_slices[0]
                .stages
            )
            == 5
        )

        # Check specific stages
        collect_stage = (
            config.pipelines[0]
            .thin_slices[0]
            .stages[0]
        )
        evolve_stage = (
            config.pipelines[0]
            .thin_slices[0]
            .stages[2]
        )

        assert (
            collect_stage.name
            == "1c_collect"
        )
        assert (
            len(collect_stage.b_units)
            == 2
        )
        assert (
            collect_stage.b_units[0]
            == "ca_b_unit"
        )

        assert (
            evolve_stage.name
            == "3e_evolve"
        )
        assert (
            len(evolve_stage.sub_stages)
            == 2
        )
        assert (
            evolve_stage.sub_stages[
                0
            ].name
            == "sub_stage_1"
        )
        assert (
            evolve_stage.sub_stages[
                1
            ].name
            == "sub_stage_2"
        )

    def test_validate_pipeline_config_missing_domain(
        self,
    ):
        """Test validation with missing domain_name."""
        invalid_config = {
            "pipelines": []
        }

        with pytest.raises(
            ValueError
        ) as exc_info:
            validate_pipeline_config(
                invalid_config
            )

        assert (
            "domain_name is required"
            in str(exc_info.value)
        )

    def test_validate_pipeline_config_missing_pipeline_name(
        self,
    ):
        """Test validation with missing pipeline name."""
        invalid_config = {
            "domain_name": "test_domain",
            "pipelines": [{}],
        }

        with pytest.raises(
            ValueError
        ) as exc_info:
            validate_pipeline_config(
                invalid_config
            )

        assert (
            "Pipeline name is required"
            in str(exc_info.value)
        )

    def test_generate_pipeline_from_config(
        self,
        pipeline_builder_configuration_file_path,
        pipeline_output_folder_absolute_path,
        template_pipeline_folder_absolute_path,
    ):
        """Test generating a pipeline from a configuration file."""
        # Load the test configuration
        with open(
            pipeline_builder_configuration_file_path,
            "r",
        ) as f:
            config_dict = json.load(f)

        # Extract configuration values for assertions
        domain_name = config_dict[
            "domain_name"
        ]
        first_pipeline_name = (
            config_dict["pipelines"][0][
                "name"
            ]
        )
        first_thin_slice = config_dict[
            "pipelines"
        ][0]["thin_slices"][0]["name"]

        # Generate the pipeline
        pipeline_path = generate_pipeline(
            config_dict,
            pipeline_output_folder_absolute_path,
        )

        # Check the pipeline directory was created
        domain_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "bclearer_pipelines",
            domain_name,
        )
        assert os.path.exists(
            domain_path
        )

        # Check key directories
        assert os.path.exists(
            os.path.join(
                domain_path, "b_source"
            )
        )
        assert os.path.exists(
            os.path.join(
                domain_path, "resources"
            )
        )
        assert os.path.exists(
            os.path.join(
                domain_path, "tests"
            )
        )

        # Check app runner file
        app_runner_path = os.path.join(
            domain_path,
            "b_source",
            "app_runners",
            f"{domain_name}_b_clearer_pipeline_b_application_runner.py",
        )
        assert os.path.exists(
            app_runner_path
        )

        # Check pipeline directories
        pipeline_path = os.path.join(
            domain_path,
            "b_source",
            first_pipeline_name,
        )
        assert os.path.exists(
            pipeline_path
        )
        assert os.path.exists(
            os.path.join(
                pipeline_path, "objects"
            )
        )
        assert os.path.exists(
            os.path.join(
                pipeline_path,
                "orchestrators",
            )
        )

        # Check pipeline orchestrator
        orchestrator_path = os.path.join(
            pipeline_path,
            "orchestrators",
            "pipeline",
            f"{first_pipeline_name}_orchestrator.py",
        )
        assert os.path.exists(
            orchestrator_path
        )

        # Check thin slice orchestrator
        thin_slice_path = os.path.join(
            pipeline_path,
            "orchestrators",
            "thin_slices",
            f"{first_thin_slice}_orchestrator.py",
        )
        assert os.path.exists(
            thin_slice_path
        )

        # Check stage orchestrator
        stages_path = os.path.join(
            pipeline_path,
            "orchestrators",
            "stages",
        )
        assert os.path.exists(
            os.path.join(
                stages_path,
                f"{first_pipeline_name}_1c_collect_orchestrator.py",
            )
        )
        assert os.path.exists(
            os.path.join(
                stages_path,
                f"{first_pipeline_name}_2l_load_orchestrator.py",
            )
        )
        assert os.path.exists(
            os.path.join(
                stages_path,
                f"{first_pipeline_name}_3e_evolve_orchestrator.py",
            )
        )

        # Check b_units
        b_units_path = os.path.join(
            pipeline_path,
            "objects",
            "b_units",
        )

        # Extract configuration values for b_units and sub-stages
        first_collect_stage_b_units = (
            config_dict["pipelines"][0][
                "thin_slices"
            ][0]["stages"][0]["b_units"]
        )
        evolve_stage_index = next(
            (
                i
                for i, stage in enumerate(
                    config_dict[
                        "pipelines"
                    ][0]["thin_slices"][
                        0
                    ][
                        "stages"
                    ]
                )
                if stage["name"]
                == "3e_evolve"
            ),
            None,
        )

        if (
            evolve_stage_index
            is not None
        ):
            evolve_stage = config_dict[
                "pipelines"
            ][0]["thin_slices"][0][
                "stages"
            ][
                evolve_stage_index
            ]
            if evolve_stage[
                "sub_stages"
            ]:
                first_sub_stage_name = (
                    evolve_stage[
                        "sub_stages"
                    ][0]["name"]
                )
                first_sub_stage_b_units = evolve_stage[
                    "sub_stages"
                ][
                    0
                ][
                    "b_units"
                ]

                # Check collect stage b_units
                collect_b_units_path = os.path.join(
                    b_units_path,
                    f"{first_pipeline_name}_1c_collect",
                )
                assert os.path.exists(
                    collect_b_units_path
                )
                for (
                    b_unit
                ) in first_collect_stage_b_units:
                    assert os.path.exists(
                        os.path.join(
                            collect_b_units_path,
                            f"{b_unit}_b_units.py",
                        )
                    )

                # Check evolve stage sub-stages
                evolve_sub_stages_path = os.path.join(
                    pipeline_path,
                    "orchestrators",
                    "sub_stages",
                    f"{first_pipeline_name}_3e_evolve_{first_sub_stage_name}",
                )
                assert os.path.exists(
                    evolve_sub_stages_path
                )
                assert os.path.exists(
                    os.path.join(
                        evolve_sub_stages_path,
                        f"{first_pipeline_name}_3e_evolve_{first_sub_stage_name}_orchestrator.py",
                    )
                )

                # Check sub-stage b_units
                evolve_sub_stage_b_units_path = os.path.join(
                    b_units_path,
                    f"{first_pipeline_name}_3e_evolve",
                    f"{first_pipeline_name}_3e_evolve_{first_sub_stage_name}",
                )
                assert os.path.exists(
                    evolve_sub_stage_b_units_path
                )
                for (
                    b_unit
                ) in first_sub_stage_b_units:
                    assert os.path.exists(
                        os.path.join(
                            evolve_sub_stage_b_units_path,
                            f"{b_unit}_b_units.py",
                        )
                    )

    def test_generate_pipeline_file_contents(
        self,
        pipeline_builder_configuration_file_path,
        pipeline_output_folder_absolute_path,
        template_pipeline_folder_absolute_path,
    ):
        """Test that generated files have the correct content."""
        # Load the test configuration
        with open(
            pipeline_builder_configuration_file_path,
            "r",
        ) as f:
            config_dict = json.load(f)

        # Extract configuration values for assertions
        domain_name = config_dict[
            "domain_name"
        ]
        first_pipeline_name = (
            config_dict["pipelines"][0][
                "name"
            ]
        )
        first_thin_slice = config_dict[
            "pipelines"
        ][0]["thin_slices"][0]["name"]
        first_b_unit = config_dict[
            "pipelines"
        ][0]["thin_slices"][0][
            "stages"
        ][
            0
        ][
            "b_units"
        ][
            0
        ]

        # Generate the pipeline
        generate_pipeline(
            config_dict,
            pipeline_output_folder_absolute_path,
        )

        # Check the application runner file content
        app_runner_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "bclearer_pipelines",
            domain_name,
            "b_source",
            "app_runners",
            f"{domain_name}_b_clearer_pipeline_b_application_runner.py",
        )

        with open(
            app_runner_path, "r"
        ) as f:
            content = f.read()
            assert (
                "from bclearer_orchestration_services.b_app_runner_service.b_application_runner import"
                in content
            )
            assert (
                f"from bclearer_pipelines.{domain_name}.b_source.app_runners.runners.{domain_name}_b_clearer_pipelines_runner import"
                in content
            )
            assert (
                f"run_{domain_name}_b_clearer_pipelines"
                in content
            )
            assert (
                f"run_{domain_name}_b_clearer_pipeline_b_application"
                in content
            )

        # Check the pipelines runner file content
        pipelines_runner_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "bclearer_pipelines",
            domain_name,
            "b_source",
            "app_runners",
            "runners",
            f"{domain_name}_b_clearer_pipelines_runner.py",
        )

        with open(
            pipelines_runner_path, "r"
        ) as f:
            content = f.read()
            assert (
                "from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper_latest import"
                in content
            )
            assert (
                f"from bclearer_pipelines.{domain_name}.b_source.app_runners.runners.{first_pipeline_name}_runner import"
                in content
            )
            assert (
                f"run_{first_pipeline_name}"
                in content
            )

        # Check b_unit file content
        b_unit_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "bclearer_pipelines",
            domain_name,
            "b_source",
            first_pipeline_name,
            "objects",
            "b_units",
            f"{first_pipeline_name}_1c_collect",
            f"{first_b_unit}_b_units.py",
        )

        with open(
            b_unit_path, "r"
        ) as f:
            content = f.read()
            assert (
                "from bclearer_core.configurations.datastructure.logging_inspection_level_b_enums import"
                in content
            )

            # Create proper class name for assertion (CaBUnitBUnits)
            class_name_parts = (
                first_b_unit.split("_")
            )
            class_name = "".join(
                part.capitalize()
                for part in class_name_parts
            )
            assert (
                f"class {class_name}BUnits:"
                in content
            )
            assert (
                "def run(self) -> None:"
                in content
            )
            assert (
                "def b_unit_process_function("
                in content
            )

        # Check orchestrator file content
        orchestrator_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "bclearer_pipelines",
            domain_name,
            "b_source",
            first_pipeline_name,
            "orchestrators",
            "pipeline",
            f"{first_pipeline_name}_orchestrator.py",
        )

        with open(
            orchestrator_path, "r"
        ) as f:
            content = f.read()
            assert (
                f"from bclearer_pipelines.{domain_name}.b_source.{first_pipeline_name}.orchestrators.thin_slices.{first_thin_slice}_orchestrator import"
                in content
            )
            assert (
                f"orchestrate_{first_thin_slice}"
                in content
            )
            assert (
                f"def orchestrate_{first_pipeline_name}():"
                in content
            )



from enum import Enum


class BoroObjectCkIds(Enum):
    NotSet = 0

    # region Top Ontology

    Objects = 1
    Types = 2
    Elements = 3
    Tuples = 4
    PlaceableObjects = 5
    PlaceableTypes = 6
    NonPlaceableTypes = 7
    TypesInstances = 8
    SuperSubTypes = 9
    WholesParts = 10
    TypePlaces = 11

    # endregion  // Top Ontology

    # region Ontology

    AcyclicCoupleTypes = 1001
    AntisymmetricCoupleTypes = 1002
    AsymmetricCoupleTypes = 1003
    CharacterStrings = 1004
    Couples = 1005
    CycleTupleTypes = 1006
    CyclicTupleTypes = 1007

    DescribedBy = 1008
    DescribingSpaces = 1009
    Descriptions = 1010
    FixedArityTupleTypes = 1011
    HomogenousFiniteCountTypeTypes = (
        1012
    )
    IntendedCharacterEquivalentStringNames = (
        1013
    )
    IntendedDistinctCharactersNamingSpaces = (
        1014
    )
    IntendedReservedCharactersNamingSpaces = (
        1015
    )
    IntendedUniqueCharactersNamingSpaces = (
        1016
    )
    IntransitiveCoupleTypes = 1017
    IrreflexiveCouples = 1018
    IrreflexiveCoupleTypes = 1019
    NamedBy = 1020
    Names = 1021
    NamesPowertype = 1022
    NamingSpaces = 1023
    NfAlternateNames = 1024
    NfBusinessDomainDescriptions = 1025
    NfCommonReservedNames = 1026
    NfEarCommonReservedNames = 1027
    NonWellFoundedTypesInstances = 1028
    NumberOfFiniteCountTypesInstancesTypes = (
        1029
    )
    PartialOrderCoupleTypes = 1030
    PowerTypesInstances = 1031
    ProximalSuperSubTypes = 1032
    ProximalTypesInstances = 1033
    ProximalWholesParts = 1034
    Quadruples = 1035
    ReflexiveCouples = 1036
    ReflexiveCoupleTypes = 1037
    RemoteSuperSubTypes = 1038
    RemoteTypesInstances = 1039
    RemoteWholesParts = 1040
    Representations = 1041
    RepresentedBy = 1042
    Signs = 1043
    Singletons = 1044
    StrictSuperSubTypes = 1045
    StrictWholesParts = 1046
    SymmetricCoupleTypes = 1047
    TemporalStages = 1048
    TransitiveCoupleTypes = 1049
    Triples = 1050
    VariableArityTupleTypes = 1051

    BoroCoreIdentifiers = 1052

    # endregion  // Ontology

    # region Agentology

    DefaultNfMentionFormulaNames = 2001
    DefaultObjectSignStoreFormulaNames = (
        2002
    )
    ObjectSignRepositoryIndices = 2003
    ThisNfAgent = 2004
    TnfaCommonReservedNames = 2005
    TnfaEarCommonReservedNames = 2006
    TnfaExemplarTypesInstances = 2007
    TnfaRootNamingSpaceNames = 2008

    # endregion  // Agentology


def get_immutable_tuple_infrastructure_subtype_ck_ids():
    immutable_tuple_infrastructure_subtype_ck_ids = {
        BoroObjectCkIds.Tuples,
        BoroObjectCkIds.SuperSubTypes,
        BoroObjectCkIds.WholesParts,
        BoroObjectCkIds.TypesInstances,
        BoroObjectCkIds.PowerTypesInstances,
        BoroObjectCkIds.NamedBy,
        BoroObjectCkIds.RepresentedBy,
    }

    return immutable_tuple_infrastructure_subtype_ck_ids


def get_composition_couple_types_ck_ids():
    composition_couple_types_ck_ids = {
        BoroObjectCkIds.SuperSubTypes,
        BoroObjectCkIds.WholesParts,
        BoroObjectCkIds.TypesInstances,
        BoroObjectCkIds.PowerTypesInstances,
    }

    return (
        composition_couple_types_ck_ids
    )


def is_immutable_tuple_infrastructure_subtype_reflexive(
    boroObjectCkId,
):
    if (
        boroObjectCkId
        == BoroObjectCkIds.SuperSubTypes
    ):
        return True

    if (
        boroObjectCkId
        == BoroObjectCkIds.WholesParts
    ):
        return True

    return False



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)


def rename_connector(
    nf_ea_com_universe: NfEaComUniverses,
    place_1_nf_uuid: str,
    place_2_nf_uuid: str,
    connector_type: EaConnectorTypes,
    connector_name: str,
) -> None:
    connectors = nf_ea_com_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.EA_CONNECTORS
    ]

    connectors_connecting_place_1 = connectors[
        connectors[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name
        ]
        == place_1_nf_uuid
    ]

    connectors_connecting_place_1_and_2 = connectors_connecting_place_1[
        connectors_connecting_place_1[
            NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name
        ]
        == place_2_nf_uuid
    ]

    connectors_connecting_place_1_and_2_of_type = connectors_connecting_place_1_and_2[
        connectors_connecting_place_1_and_2[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == connector_type.type_name
    ]

    connector_nf_uuid = list(
        connectors_connecting_place_1_and_2_of_type[
            NfColumnTypes.NF_UUIDS.column_name
        ],
    )[0]

    connectors.loc[
        connectors[
            NfColumnTypes.NF_UUIDS.column_name
        ]
        == connector_nf_uuid,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name,
    ] = connector_name

    nf_ea_com_universe.nf_ea_com_registry.replace_collection(
        collection_type=NfEaComCollectionTypes.EA_CONNECTORS,
        collection=connectors,
    )



from bclearer_core.common_knowledge.bclearer_additional_column_types import (
    BclearerAdditionalColumnTypes,
)
from bclearer_core.common_knowledge.bclearer_constants import (
    NAME_INSTANCE_ATTRIBUTE_NAME,
)
from bclearer_core.common_knowledge.bclearer_matched_ea_objects import (
    BclearerMatchedEaObjects,
)
from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.constants.standard_constants import (
    DEFAULT_FOREIGN_TABLE_SUFFIX,
    DEFAULT_MASTER_TABLE_SUFFIX,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.common.new_root_package_creator import (
    create_root_package,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.ea_attribute_remover import (
    remove_ea_attributes,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.name_instance_creator import (
    create_name_instance,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.name_instance_type_creator import (
    create_name_instance_type,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_interop_services.dataframe_service.dataframe_helpers.dataframe_mergers import (
    left_merge_dataframes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import DataFrame


def shift_convention_separate_standard_names_and_instances(
    content_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    package_name: str,
) -> NfEaComUniverses:
    log_message(
        message="Separate Standard Names and Instances - started",
    )

    output_universe.nf_ea_com_registry.dictionary_of_collections = (
        content_universe.nf_ea_com_registry.dictionary_of_collections.copy()
    )

    __separate_names_and_instances(
        nf_ea_com_universe=output_universe,
        package_name=package_name,
    )

    log_message(
        message="Separate Standard Names and Instances - finished",
    )

    return output_universe


def __separate_names_and_instances(
    nf_ea_com_universe: NfEaComUniverses,
    package_name: str,
):
    package_nf_uuid = create_root_package(
        nf_ea_com_universe=nf_ea_com_universe,
        package_name=package_name,
    )

    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    attributes_grouped_by_naming_space_dictionary = __get_attributes_grouped_by_naming_space_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
    )

    digitalisation_level_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE.ea_guid,
    )

    for (
        naming_space_nf_uuid,
        attributes,
    ) in (
        attributes_grouped_by_naming_space_dictionary.items()
    ):
        __separate_naming_space_instances(
            nf_ea_com_universe=nf_ea_com_universe,
            naming_space_nf_uuid=naming_space_nf_uuid,
            ea_attributes=attributes,
            new_ea_objects_dictionary=new_ea_objects_dictionary,
            package_nf_uuid=package_nf_uuid,
            digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )


def __separate_naming_space_instances(
    nf_ea_com_universe: NfEaComUniverses,
    naming_space_nf_uuid: str,
    ea_attributes: DataFrame,
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    digitalisation_level_stereotype_nf_uuid: str,
):
    ea_classifiers = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    naming_space_name = ea_classifiers.at[
        ea_classifiers[
            NfColumnTypes.NF_UUIDS.column_name
        ]
        .eq(naming_space_nf_uuid)
        .idxmax(),
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name,
    ]

    name_instance_type_name = (
        naming_space_name[:-1]
        + " Instances"
    )

    name_instance_type_nf_uuid = create_name_instance_type(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
        package_nf_uuid=package_nf_uuid,
        name_object_name=name_instance_type_name,
        naming_space_nf_uuid=naming_space_nf_uuid,
        digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
    )

    name_types_instances_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=BclearerMatchedEaObjects.NAME_TYPES_INSTANCES_STEREOTYPE.ea_guid,
    )

    for (
        ea_attribute_tuple
    ) in ea_attributes.itertuples():
        create_name_instance(
            new_ea_objects_dictionary=new_ea_objects_dictionary,
            package_nf_uuid=package_nf_uuid,
            ea_attribute_tuple=ea_attribute_tuple,
            name_instance_type_nf_uuid=name_instance_type_nf_uuid,
            name_types_instances_stereotype_nf_uuid=name_types_instances_stereotype_nf_uuid,
            digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
        )

    remove_ea_attributes(
        nf_ea_com_universe=nf_ea_com_universe,
        ea_attributes=ea_attributes,
    )


def __get_attributes_grouped_by_naming_space_dictionary(
    nf_ea_com_universe: NfEaComUniverses,
) -> dict:
    ea_attributes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_attributes()
    )

    standard_name_instance_attributes = ea_attributes[
        ea_attributes[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
        ]
        == NAME_INSTANCE_ATTRIBUTE_NAME
    ]

    ea_connectors = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_connectors()
    )

    ea_dependencies = ea_connectors[
        ea_connectors[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == EaConnectorTypes.DEPENDENCY.type_name
    ]

    extended_standard_name_instance_attributes = left_merge_dataframes(
        master_dataframe=standard_name_instance_attributes,
        master_dataframe_key_columns=[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
        ],
        merge_suffixes=[
            DEFAULT_FOREIGN_TABLE_SUFFIX,
            DEFAULT_MASTER_TABLE_SUFFIX,
        ],
        foreign_key_dataframe=ea_dependencies,
        foreign_key_dataframe_fk_columns=[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name,
        ],
        foreign_key_dataframe_other_column_rename_dictionary={
            NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name: BclearerAdditionalColumnTypes.NAMING_SPACE_NF_UUIDS.column_name,
        },
    )

    ea_classifiers = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    extended_standard_name_instance_attributes = left_merge_dataframes(
        master_dataframe=extended_standard_name_instance_attributes,
        master_dataframe_key_columns=[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
        ],
        merge_suffixes=[
            DEFAULT_FOREIGN_TABLE_SUFFIX,
            DEFAULT_MASTER_TABLE_SUFFIX,
        ],
        foreign_key_dataframe=ea_classifiers,
        foreign_key_dataframe_fk_columns=[
            NfColumnTypes.NF_UUIDS.column_name,
        ],
        foreign_key_dataframe_other_column_rename_dictionary={
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: BclearerAdditionalColumnTypes.OWNING_OBJECT_NAMES.column_name,
        },
    )

    naming_space_nf_uuids = set(
        extended_standard_name_instance_attributes[
            BclearerAdditionalColumnTypes.NAMING_SPACE_NF_UUIDS.column_name
        ],
    )

    attributes_grouped_by_name_space_dictionary = (
        {}
    )

    for (
        naming_space_nf_uuid
    ) in naming_space_nf_uuids:
        naming_space_ea_attributes = extended_standard_name_instance_attributes[
            extended_standard_name_instance_attributes[
                BclearerAdditionalColumnTypes.NAMING_SPACE_NF_UUIDS.column_name
            ]
            == naming_space_nf_uuid
        ]

        attributes_grouped_by_name_space_dictionary[
            naming_space_nf_uuid
        ] = naming_space_ea_attributes

    return attributes_grouped_by_name_space_dictionary



"""Pipeline generator module for creating pipeline structure from configuration."""

# Import all functionality from the more modular template-based generator
from bclearer_core.pipeline_builder.generator_with_templates import (
    PipelineGenerator,
    PipelineUpdater,
    generate_pipeline,
    update_pipeline,
)



from bclearer_core.common_knowledge.universe_modification_operation_types import (
    UniverseModificationOperationTypes,
)
from bclearer_core.configurations.operation_configurations import (
    OperationConfigurations,
)
from bclearer_core.configurations.universe_modification_configuration_objects import (
    UniverseModificationConfigurationObjects,
)


class UniverseModificationOperationConfigurations(
    OperationConfigurations,
):
    def __init__(
        self,
        universe_modification_operation_type: UniverseModificationOperationTypes,
        output_universe_short_name: str,
        universe_modification_configuration_object: UniverseModificationConfigurationObjects = None,
        package_name: str = None,
    ):
        super().__init__(
            operation_type=universe_modification_operation_type,
        )

        self.output_universe_short_name = (
            output_universe_short_name
        )

        self.universe_modification_configuration_object = universe_modification_configuration_object

        self.package_name = package_name

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



import os

import pytest


@pytest.fixture(scope="session")
def data_folder_absolute_path():
    data_input_relative_path = "../data"
    base_path = os.path.dirname(
        os.path.abspath(__file__),
    )
    data_folder_absolute_path = os.path.normpath(
        os.path.join(
            base_path,
            data_input_relative_path,
        ),
    )
    return data_folder_absolute_path


@pytest.fixture(scope="session")
def data_input_folder_absolute_path(
    data_folder_absolute_path,
):
    data_input_relative_path = "input"

    data_input_folder_absolute_path = (
        os.path.join(
            data_folder_absolute_path,
            data_input_relative_path,
        )
    )

    return (
        data_input_folder_absolute_path
    )


@pytest.fixture(scope="session")
def data_output_folder_absolute_path(
    data_folder_absolute_path,
):
    data_output_relative_path = "output"

    data_output_folder_absolute_path = (
        os.path.join(
            data_folder_absolute_path,
            data_output_relative_path,
        )
    )

    return (
        data_output_folder_absolute_path
    )


@pytest.fixture(scope="session")
def configurations_folder_absolute_path():
    configurations_folder_relative_path = (
        "../configurations"
    )
    base_path = os.path.dirname(
        os.path.abspath(__file__),
    )
    configurations_folder_absolute_path = os.path.normpath(
        os.path.join(
            base_path,
            configurations_folder_relative_path,
        ),
    )
    return configurations_folder_absolute_path


@pytest.fixture(scope="session")
def log_folder_absolute_path(
    data_folder_absolute_path,
):
    log_folder_relative_path = "logs"
    log_folder_absolute_path = (
        os.path.join(
            data_folder_absolute_path,
            log_folder_relative_path,
        )
    )
    return log_folder_absolute_path


@pytest.fixture(scope="session")
def template_pipeline_folder_absolute_path():
    """
    Find the template_pipeline directory by searching up from the current directory.
    This approach is more robust to different directory structures.
    """
    # Start from the current directory
    current_dir = os.path.dirname(
        os.path.abspath(__file__)
    )

    # Try different possible paths to find the template_pipeline directory
    possible_relative_paths = [
        "../../../../pipelines/template_pipeline",  # If 4 levels up from fixtures
        "../../../pipelines/template_pipeline",  # If 3 levels up from fixtures
        "../../../../../pipelines/template_pipeline",  # If 5 levels up from fixtures
        "../../../../../../pipelines/template_pipeline",  # If 6 levels up from fixtures
    ]

    for (
        relative_path
    ) in possible_relative_paths:
        path = os.path.normpath(
            os.path.join(
                current_dir,
                relative_path,
            )
        )
        if os.path.exists(
            path
        ) and os.path.isdir(path):
            return path

    # If we're here, we need to search upwards through the directory tree
    root_dir = current_dir
    while True:
        # Move up one level
        parent_dir = os.path.dirname(
            root_dir
        )
        if (
            parent_dir == root_dir
        ):  # We've reached the filesystem root
            break
        root_dir = parent_dir

        # Check if pipelines/template_pipeline exists at this level
        template_path = os.path.join(
            root_dir,
            "pipelines",
            "template_pipeline",
        )
        if os.path.exists(
            template_path
        ) and os.path.isdir(
            template_path
        ):
            return template_path

    # If we couldn't find it, try using a direct path from D: drive (for Windows systems)
    if os.name == "nt":
        direct_path = r"D:\S\bclearer\ol_bclearer_pdk\pipelines\template_pipeline"
        if os.path.exists(
            direct_path
        ) and os.path.isdir(
            direct_path
        ):
            return direct_path

    # If we get here, we couldn't find the template pipeline directory
    raise FileNotFoundError(
        "Could not find template_pipeline directory. Please make sure it exists "
        "in your repository at 'pipelines/template_pipeline'."
    )


@pytest.fixture(scope="session")
def pipeline_builder_configuration_file_path(
    configurations_folder_absolute_path,
):
    config_path_relative = "pipeline_builder/sample_pipeline_config.json"
    config_file_path = os.path.join(
        configurations_folder_absolute_path,
        config_path_relative,
    )
    return config_file_path


@pytest.fixture(scope="session")
def pipeline_output_folder_absolute_path(
    data_output_folder_absolute_path,
):
    output_folder_relative = (
        "pipeline_builder"
    )
    output_folder_path = os.path.join(
        data_output_folder_absolute_path,
        output_folder_relative,
    )

    # Create the folder if it doesn't exist
    os.makedirs(
        output_folder_path,
        exist_ok=True,
    )

    return output_folder_path



from enum import auto, unique

from bclearer_core.nf.types.collection_types import (
    CollectionTypes,
)


@unique
class CommonCollectionTypes(
    CollectionTypes,
):
    SUMMARY_TABLE = auto()

    def __collection_name(self) -> str:
        collection_name = (
            collection_name_mapping[
                self
            ]
        )

        return collection_name

    collection_name = property(
        fget=__collection_name,
    )


collection_name_mapping = {
    CommonCollectionTypes.SUMMARY_TABLE: "summary_table",
}



from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)


def get_instances_nf_uuids_of_matched_type(
    nf_ea_com_universe: NfEaComUniverses,
    matched_type: MatchedEaObjects,
) -> set:
    matched_type_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=matched_type.ea_guid,
    )

    instances_nf_uuids = get_instances_nf_uuids_of_type_nf_uuid(
        nf_ea_com_universe=nf_ea_com_universe,
        nf_uuid=matched_type_nf_uuid,
    )

    return instances_nf_uuids


def get_instances_nf_uuids_of_type_nf_uuid(
    nf_ea_com_universe: NfEaComUniverses,
    nf_uuid: str,
) -> set:
    ea_connectors = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_connectors()
    )

    ea_dependencies = ea_connectors[
        ea_connectors[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == EaConnectorTypes.DEPENDENCY.type_name
    ]

    filtered_dependencies = ea_dependencies[
        ea_dependencies[
            NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name
        ]
        == nf_uuid
    ]

    instances_nf_uuids = set(
        filtered_dependencies[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name
        ],
    )

    return instances_nf_uuids



"""Utility to extract templates from the pipeline template directory."""

import os
import re
from pathlib import Path


class TemplateExtractor:
    """Extracts templates from the pipeline template directory and updates the constants."""

    def __init__(
        self, template_path: str
    ):
        """
        Initialize the TemplateExtractor.

        Args:
            template_path: Path to the template pipeline directory
        """
        self.template_path = Path(
            template_path
        )

        if (
            not self.template_path.exists()
        ):
            raise FileNotFoundError(
                f"Template path not found: {template_path}"
            )

        if (
            not self.template_path.is_dir()
        ):
            raise NotADirectoryError(
                f"Template path is not a directory: {template_path}"
            )

    def extract_templates(
        self, output_directory: str
    ) -> None:
        """
        Extract templates from the template pipeline and save them to output directory.

        Args:
            output_directory: Path to save the extracted templates
        """
        output_path = Path(
            output_directory
        )
        output_path.mkdir(
            parents=True, exist_ok=True
        )

        # Extract application runner
        self._extract_application_runner(
            output_path
        )

        # Extract pipelines runner
        self._extract_pipelines_runner(
            output_path
        )

        # Extract pipeline runner
        self._extract_pipeline_runner(
            output_path
        )

        # Extract pipeline orchestrator
        self._extract_pipeline_orchestrator(
            output_path
        )

        # Extract thin slice orchestrator
        self._extract_thin_slice_orchestrator(
            output_path
        )

        # Extract stage orchestrator
        self._extract_stage_orchestrator(
            output_path
        )

        # Extract sub-stage orchestrator
        self._extract_sub_stage_orchestrator(
            output_path
        )

        # Extract b_unit
        self._extract_b_unit(
            output_path
        )

        # Extract b_unit_creator_and_runner
        self._extract_b_unit_creator_and_runner(
            output_path
        )

        # Extract conftest
        self._extract_conftest(
            output_path
        )

        # Extract e2e test
        self._extract_e2e_test(
            output_path
        )

        # Create __init__.py
        self._create_init(output_path)

    def _extract_application_runner(
        self, output_path: Path
    ) -> None:
        """
        Extract application runner template.

        Args:
            output_path: Path to save the template
        """
        template_file = (
            self.template_path
            / "b_source"
            / "app_runners"
            / "aa_b_clearer_pipeline_b_application_runner.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Application runner template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Replace domain-specific names with placeholders
        content = content.replace(
            "aa", "{domain_name}"
        )

        # Save to output file
        output_file = (
            output_path
            / "application_runner.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for application runner."""\n\n'
            )
            f.write(
                f'APPLICATION_RUNNER_TEMPLATE = """{content}"""'
            )

    def _extract_pipelines_runner(
        self, output_path: Path
    ) -> None:
        """
        Extract pipelines runner template.

        Args:
            output_path: Path to save the template
        """
        template_file = (
            self.template_path
            / "b_source"
            / "app_runners"
            / "runners"
            / "aa_b_clearer_pipelines_runner.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Pipelines runner template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Extract header (imports)
        header_match = re.search(
            r"^.*?import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if header_match:
            header = header_match.group(
                0
            )
        else:
            header = ""

        # Extract pipeline import pattern
        import_match = re.search(
            r"^from.*?xx_runner.*?import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if import_match:
            pipeline_import = (
                import_match.group(0)
            )
            pipeline_import = (
                pipeline_import.replace(
                    "aa",
                    "{domain_name}",
                )
            )
            pipeline_import = (
                pipeline_import.replace(
                    "xx",
                    "{pipeline_name}",
                )
            )
        else:
            pipeline_import = ""

        # Extract function start
        function_start_match = re.search(
            r"@run_and_log_function\(\).*?def run_aa_b_clearer_pipelines.*?:.*?\n",
            content,
            re.DOTALL,
        )
        if function_start_match:
            function_start = function_start_match.group(
                0
            )
            function_start = (
                function_start.replace(
                    "aa",
                    "{domain_name}",
                )
            )
        else:
            function_start = ""

        # Extract pipeline call
        pipeline_call_match = re.search(
            r"^\s+run_xx\(\)",
            content,
            re.MULTILINE,
        )
        if pipeline_call_match:
            pipeline_call = pipeline_call_match.group(
                0
            )
            pipeline_call = (
                pipeline_call.replace(
                    "xx",
                    "{pipeline_name}",
                )
            )
        else:
            pipeline_call = "    run_{pipeline_name}()"

        # Extract empty function body if it exists
        empty_body_match = re.search(
            r"^\s+pass$",
            content,
            re.MULTILINE,
        )
        if empty_body_match:
            empty_body = (
                empty_body_match.group(
                    0
                )
            )
        else:
            empty_body = "    pass"

        # Save to output file
        output_file = (
            output_path
            / "pipelines_runner.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for pipelines runner."""\n\n'
            )
            f.write(
                f'PIPELINES_RUNNER_TEMPLATE_HEADER = """{header}"""\n\n'
            )
            f.write(
                f'PIPELINE_IMPORT_TEMPLATE = """{pipeline_import}"""\n\n'
            )
            f.write(
                f'PIPELINES_RUNNER_FUNCTION_START = """{function_start}"""\n\n'
            )
            f.write(
                f'PIPELINE_RUNNER_CALL = """{pipeline_call}\n"""\n\n'
            )
            f.write(
                f'EMPTY_FUNCTION_BODY = """{empty_body}\n"""'
            )

    # Add the remaining extraction methods here following the same pattern
    def _extract_pipeline_runner(
        self, output_path: Path
    ) -> None:
        """Extract pipeline runner template."""
        template_file = (
            self.template_path
            / "b_source"
            / "app_runners"
            / "runners"
            / "xx_runner.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Pipeline runner template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Replace domain-specific names with placeholders
        content = content.replace(
            "aa", "{domain_name}"
        )
        content = content.replace(
            "xx", "{pipeline_name}"
        )

        # Save to output file
        output_file = (
            output_path
            / "pipeline_runner.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for pipeline runner."""\n\n'
            )
            f.write(
                f'PIPELINE_RUNNER_TEMPLATE = """{content}"""'
            )

    def _extract_pipeline_orchestrator(
        self, output_path: Path
    ) -> None:
        """Extract pipeline orchestrator template."""
        template_file = (
            self.template_path
            / "b_source"
            / "xx_pipeline"
            / "orchestrators"
            / "pipeline"
            / "xx_pipeline_orchestrator.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Pipeline orchestrator template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Extract import pattern
        import_match = re.search(
            r"^from.*?thin_slices.*?import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if import_match:
            import_pattern = (
                import_match.group(0)
            )
            import_pattern = (
                import_pattern.replace(
                    "aa",
                    "{domain_name}",
                )
            )
            import_pattern = (
                import_pattern.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            import_pattern = (
                import_pattern.replace(
                    "thin_slice_1",
                    "{thin_slice_name}",
                )
            )
        else:
            import_pattern = ""

        # Extract function start
        function_start_match = re.search(
            r"def orchestrate_xx_pipeline.*?__run_contained_bie_pipeline_components.*?def __run_contained_bie_pipeline_components.*?:.*?\n",
            content,
            re.DOTALL,
        )
        if function_start_match:
            function_start = function_start_match.group(
                0
            )
            function_start = (
                function_start.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
        else:
            function_start = ""

        # Extract thin slice call
        thin_slice_call_match = re.search(
            r"^\s+orchestrate_thin_slice_1\(\)",
            content,
            re.MULTILINE,
        )
        if thin_slice_call_match:
            thin_slice_call = thin_slice_call_match.group(
                0
            )
            thin_slice_call = (
                thin_slice_call.replace(
                    "thin_slice_1",
                    "{thin_slice_name}",
                )
            )
        else:
            thin_slice_call = "    orchestrate_{thin_slice_name}()"

        # Extract empty function body if it exists
        empty_body_match = re.search(
            r"^\s+pass$",
            content,
            re.MULTILINE,
        )
        if empty_body_match:
            empty_body = (
                empty_body_match.group(
                    0
                )
            )
        else:
            empty_body = "    pass"

        # Save to output file
        output_file = (
            output_path
            / "pipeline_orchestrator.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for pipeline orchestrator."""\n\n'
            )
            f.write(
                f'PIPELINE_ORCHESTRATOR_IMPORT_TEMPLATE = """{import_pattern}"""\n\n'
            )
            f.write(
                f'PIPELINE_ORCHESTRATOR_FUNCTION_START = """{function_start}"""\n\n'
            )
            f.write(
                f'ORCHESTRATE_THIN_SLICE_CALL = """{thin_slice_call}\n"""\n\n'
            )
            f.write(
                f'EMPTY_FUNCTION_BODY = """{empty_body}\n"""'
            )

    def _extract_thin_slice_orchestrator(
        self, output_path: Path
    ) -> None:
        """Extract thin slice orchestrator template."""
        template_file = (
            self.template_path
            / "b_source"
            / "xx_pipeline"
            / "orchestrators"
            / "thin_slices"
            / "thin_slice_1_orchestrator.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Thin slice orchestrator template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Extract import pattern
        import_match = re.search(
            r"^from.*?stages.*?import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if import_match:
            import_pattern = (
                import_match.group(0)
            )
            import_pattern = (
                import_pattern.replace(
                    "aa",
                    "{domain_name}",
                )
            )
            import_pattern = (
                import_pattern.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            if (
                "2l_load"
                in import_pattern
            ):
                import_pattern = import_pattern.replace(
                    "2l_load",
                    "{stage_name}",
                )
            else:
                import_pattern = import_pattern.replace(
                    "1c_collect",
                    "{stage_name}",
                )
        else:
            import_pattern = ""

        # Extract function start
        function_start_match = re.search(
            r"def orchestrate_thin_slice_1.*?__run_contained_bie_pipeline_components.*?def __run_contained_bie_pipeline_components.*?:.*?\n",
            content,
            re.DOTALL,
        )
        if function_start_match:
            function_start = function_start_match.group(
                0
            )
            function_start = (
                function_start.replace(
                    "thin_slice_1",
                    "{thin_slice_name}",
                )
            )
        else:
            function_start = ""

        # Extract stage call
        stage_call_match = re.search(
            r"^\s+orchestrate_xx_pipeline_[12][lc]_\w+\(\)",
            content,
            re.MULTILINE,
        )
        if stage_call_match:
            stage_call = (
                stage_call_match.group(
                    0
                )
            )
            stage_call = (
                stage_call.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            stage_call = re.sub(
                r"_[12][lc]_\w+",
                "_{stage_name}",
                stage_call,
            )
        else:
            stage_call = "    orchestrate_{pipeline_name}_{stage_name}()"

        # Extract empty function body if it exists
        empty_body_match = re.search(
            r"^\s+pass$",
            content,
            re.MULTILINE,
        )
        if empty_body_match:
            empty_body = (
                empty_body_match.group(
                    0
                )
            )
        else:
            empty_body = "    pass"

        # Save to output file
        output_file = (
            output_path
            / "thin_slice_orchestrator.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for thin slice orchestrator."""\n\n'
            )
            f.write(
                f'THIN_SLICE_ORCHESTRATOR_IMPORT_TEMPLATE = """{import_pattern}"""\n\n'
            )
            f.write(
                f'THIN_SLICE_ORCHESTRATOR_FUNCTION_START = """{function_start}"""\n\n'
            )
            f.write(
                f'ORCHESTRATE_STAGE_CALL = """{stage_call}\n"""\n\n'
            )
            f.write(
                f'EMPTY_FUNCTION_BODY = """{empty_body}\n"""'
            )

    def _extract_stage_orchestrator(
        self, output_path: Path
    ) -> None:
        """Extract stage orchestrator template."""
        template_file = (
            self.template_path
            / "b_source"
            / "xx_pipeline"
            / "orchestrators"
            / "stages"
            / "xx_pipeline_2l_load_orchestrator.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Stage orchestrator template file not found: {template_file}"
            )
            # Try alternative
            template_file = (
                self.template_path
                / "b_source"
                / "xx_pipeline"
                / "orchestrators"
                / "stages"
                / "xx_pipeline_3e_evolve_orchestrator.py"
            )
            if (
                not template_file.exists()
            ):
                print(
                    f"Warning: Alternative stage orchestrator template file not found: {template_file}"
                )
                return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Extract header
        header_match = re.search(
            r"^from.*?run_and_log_function.*?import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if header_match:
            header = header_match.group(
                0
            )
        else:
            header = ""

        # Extract sub_stage import if exists
        sub_stage_import_match = re.search(
            r"^from.*?sub_stages.*?orchestrator import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if sub_stage_import_match:
            sub_stage_import = sub_stage_import_match.group(
                0
            )
            sub_stage_import = sub_stage_import.replace(
                "aa", "{domain_name}"
            )
            sub_stage_import = sub_stage_import.replace(
                "xx_pipeline",
                "{pipeline_name}",
            )
            sub_stage_import = re.sub(
                r"_[123][elc]_\w+",
                "_{stage_name}",
                sub_stage_import,
            )
            sub_stage_import = re.sub(
                r"_[a-z0-9_]+_sub_stage",
                "_{sub_stage_name}",
                sub_stage_import,
            )
        else:
            sub_stage_import = ""

        # Extract b_unit creator import if exists
        b_unit_creator_import_match = re.search(
            r"^from.*?b_units\.b_unit_creator_and_runner import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if b_unit_creator_import_match:
            b_unit_creator_import = b_unit_creator_import_match.group(
                0
            )
            b_unit_creator_import = b_unit_creator_import.replace(
                "aa", "{domain_name}"
            )
        else:
            b_unit_creator_import = ""

        # Extract b_unit import if exists
        b_unit_import_match = re.search(
            r"^from.*?b_units import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if (
            b_unit_import_match
            and "objects.b_units"
            in b_unit_import_match.group(
                0
            )
        ):
            b_unit_import = b_unit_import_match.group(
                0
            )
            b_unit_import = (
                b_unit_import.replace(
                    "aa",
                    "{domain_name}",
                )
            )
            b_unit_import = (
                b_unit_import.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            b_unit_import = re.sub(
                r"_[123][elc]_\w+",
                "_{stage_name}",
                b_unit_import,
            )
            # Extract pattern for b_unit import
            unit_match = re.search(
                r"\/([a-z0-9_]+)_b_units.py",
                b_unit_import,
            )
            if unit_match:
                unit_name = (
                    unit_match.group(1)
                )
                b_unit_import = b_unit_import.replace(
                    f"{unit_name}_b_units",
                    "{b_unit_lower}_b_units",
                )

            # Handle class name
            class_match = re.search(
                r"([A-Za-z0-9]+)BUnits",
                b_unit_import,
            )
            if class_match:
                class_name = (
                    class_match.group(1)
                )
                b_unit_import = b_unit_import.replace(
                    f"{class_name}BUnits",
                    "{b_unit_class}BUnits",
                )
        else:
            b_unit_import = ""

        # Extract function start
        function_start_match = re.search(
            r"@run_and_log_function\(\).*?def orchestrate_xx_pipeline_[123][elc]_\w+.*?__run_contained_bie_pipeline_components.*?def __run_contained_bie_pipeline_components.*?:.*?\n",
            content,
            re.DOTALL,
        )
        if function_start_match:
            function_start = function_start_match.group(
                0
            )
            function_start = (
                function_start.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            function_start = re.sub(
                r"_[123][elc]_\w+\(\)",
                "_{stage_name}()",
                function_start,
            )
        else:
            function_start = ""

        # Extract sub stage call if exists
        sub_stage_call_match = re.search(
            r"^\s+orchestrate_xx_pipeline_[123][elc]_[a-z0-9_]+_[a-z0-9_]+\(\)",
            content,
            re.MULTILINE,
        )
        if sub_stage_call_match:
            sub_stage_call = sub_stage_call_match.group(
                0
            )
            sub_stage_call = (
                sub_stage_call.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            sub_stage_call = re.sub(
                r"_[123][elc]_\w+_",
                "_{stage_name}_",
                sub_stage_call,
            )
            sub_stage_call = re.sub(
                r"_[a-z0-9_]+\(\)",
                "_{sub_stage_name}()",
                sub_stage_call,
            )
        else:
            sub_stage_call = "    orchestrate_{pipeline_name}_{stage_name}_{sub_stage_name}()"

        # Extract create_and_run_b_unit call if exists
        create_run_match = re.search(
            r"^\s+create_and_run_b_unit\(.*?b_unit_type=[A-Za-z0-9]+BUnits.*?\)",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if create_run_match:
            create_run = (
                create_run_match.group(
                    0
                )
            )
            class_match = re.search(
                r"b_unit_type=([A-Za-z0-9]+)BUnits",
                create_run,
            )
            if class_match:
                class_name = (
                    class_match.group(1)
                )
                create_run = create_run.replace(
                    f"{class_name}BUnits",
                    "{b_unit_class}BUnits",
                )
        else:
            create_run = """    create_and_run_b_unit(
        b_unit_type={b_unit_class}BUnits
    )"""

        # Extract empty function body if it exists
        empty_body_match = re.search(
            r"^\s+pass$",
            content,
            re.MULTILINE,
        )
        if empty_body_match:
            empty_body = (
                empty_body_match.group(
                    0
                )
            )
        else:
            empty_body = "    pass"

        # Save to output file
        output_file = (
            output_path
            / "stage_orchestrator.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for stage orchestrator."""\n\n'
            )
            f.write(
                f'STAGE_ORCHESTRATOR_HEADER = """{header}"""\n\n'
            )
            f.write(
                f'SUB_STAGE_IMPORT_TEMPLATE = """{sub_stage_import}"""\n\n'
            )
            f.write(
                f'B_UNIT_CREATOR_IMPORT = """{b_unit_creator_import}"""\n\n'
            )
            f.write(
                f'B_UNIT_IMPORT_TEMPLATE = """{b_unit_import}"""\n\n'
            )
            f.write(
                f'STAGE_ORCHESTRATOR_FUNCTION_START = """{function_start}"""\n\n'
            )
            f.write(
                f'ORCHESTRATE_SUB_STAGE_CALL = """{sub_stage_call}\n"""\n\n'
            )
            f.write(
                f'CREATE_AND_RUN_B_UNIT_CALL = """{create_run}\n"""\n\n'
            )
            f.write(
                f'EMPTY_FUNCTION_BODY = """{empty_body}\n"""'
            )

    def _extract_sub_stage_orchestrator(
        self, output_path: Path
    ) -> None:
        """Extract sub-stage orchestrator template."""
        template_file = (
            self.template_path
            / "b_source"
            / "xx_pipeline"
            / "orchestrators"
            / "sub_stages"
            / "xx_pipeline_3e_evolve_1_sub_stage_name"
            / "xx_pipeline_3e_evolve_1_sub_stage_name_orchestrator.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Sub-stage orchestrator template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Extract header
        header_match = re.search(
            r"^from.*?run_and_log_function.*?import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if header_match:
            header = header_match.group(
                0
            )
        else:
            header = ""

        # Extract b_unit creator import if exists
        b_unit_creator_import_match = re.search(
            r"^from.*?b_units\.b_unit_creator_and_runner import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if b_unit_creator_import_match:
            b_unit_creator_import = b_unit_creator_import_match.group(
                0
            )
            b_unit_creator_import = b_unit_creator_import.replace(
                "aa", "{domain_name}"
            )
        else:
            b_unit_creator_import = ""

        # Extract b_unit import if exists
        b_unit_import_match = re.search(
            r"^from.*?b_units import \(.*?\)$",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if (
            b_unit_import_match
            and "objects.b_units"
            in b_unit_import_match.group(
                0
            )
        ):
            b_unit_import = b_unit_import_match.group(
                0
            )
            b_unit_import = (
                b_unit_import.replace(
                    "aa",
                    "{domain_name}",
                )
            )
            b_unit_import = (
                b_unit_import.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            b_unit_import = re.sub(
                r"_[123][elc]_\w+",
                "_{stage_name}",
                b_unit_import,
            )
            b_unit_import = re.sub(
                r"_[0-9]_[a-z0-9_]+",
                "_{sub_stage_name}",
                b_unit_import,
            )

            # Extract pattern for b_unit import
            unit_match = re.search(
                r"\/([a-z0-9_]+)_b_units.py",
                b_unit_import,
            )
            if unit_match:
                unit_name = (
                    unit_match.group(1)
                )
                b_unit_import = b_unit_import.replace(
                    f"{unit_name}_b_units",
                    "{b_unit_lower}_b_units",
                )

            # Handle class name
            class_match = re.search(
                r"([A-Za-z0-9]+)BUnits",
                b_unit_import,
            )
            if class_match:
                class_name = (
                    class_match.group(1)
                )
                b_unit_import = b_unit_import.replace(
                    f"{class_name}BUnits",
                    "{b_unit_class}BUnits",
                )
        else:
            b_unit_import = ""

        # Extract function start
        function_start_match = re.search(
            r"@run_and_log_function\(\).*?def orchestrate_xx_pipeline_3e_evolve_1_sub_stage_name.*?__run_contained_bie_pipeline_components.*?def __run_contained_bie_pipeline_components.*?:.*?\n",
            content,
            re.DOTALL,
        )
        if function_start_match:
            function_start = function_start_match.group(
                0
            )
            function_start = (
                function_start.replace(
                    "xx_pipeline",
                    "{pipeline_name}",
                )
            )
            function_start = re.sub(
                r"_[123][elc]_\w+_",
                "_{stage_name}_",
                function_start,
            )
            function_start = re.sub(
                r"_[0-9]_[a-z0-9_]+\(",
                "_{sub_stage_name}(",
                function_start,
            )
        else:
            function_start = ""

        # Extract create_and_run_b_unit call if exists
        create_run_match = re.search(
            r"^\s+create_and_run_b_unit\(.*?b_unit_type=[A-Za-z0-9]+BUnits.*?\)",
            content,
            re.DOTALL | re.MULTILINE,
        )
        if create_run_match:
            create_run = (
                create_run_match.group(
                    0
                )
            )
            class_match = re.search(
                r"b_unit_type=([A-Za-z0-9]+)BUnits",
                create_run,
            )
            if class_match:
                class_name = (
                    class_match.group(1)
                )
                create_run = create_run.replace(
                    f"{class_name}BUnits",
                    "{b_unit_class}BUnits",
                )
        else:
            create_run = """    create_and_run_b_unit(
        b_unit_type={b_unit_class}BUnits
    )"""

        # Extract empty function body if it exists
        empty_body_match = re.search(
            r"^\s+pass$",
            content,
            re.MULTILINE,
        )
        if empty_body_match:
            empty_body = (
                empty_body_match.group(
                    0
                )
            )
        else:
            empty_body = "    pass"

        # Save to output file
        output_file = (
            output_path
            / "sub_stage_orchestrator.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for sub-stage orchestrator."""\n\n'
            )
            f.write(
                f'SUB_STAGE_ORCHESTRATOR_HEADER = """{header}"""\n\n'
            )
            f.write(
                f'B_UNIT_CREATOR_IMPORT = """{b_unit_creator_import}"""\n\n'
            )
            f.write(
                f'SUB_STAGE_B_UNIT_IMPORT_TEMPLATE = """{b_unit_import}"""\n\n'
            )
            f.write(
                f'SUB_STAGE_ORCHESTRATOR_FUNCTION_START = """{function_start}"""\n\n'
            )
            f.write(
                f'CREATE_AND_RUN_B_UNIT_CALL = """{create_run}\n"""\n\n'
            )
            f.write(
                f'EMPTY_FUNCTION_BODY = """{empty_body}\n"""'
            )

    def _extract_b_unit(
        self, output_path: Path
    ) -> None:
        """Extract b_unit template."""
        template_file = (
            self.template_path
            / "b_source"
            / "xx_pipeline"
            / "objects"
            / "b_units"
            / "xx_pipeline_1c_collect"
            / "ca_b_unit_b_units.py"
        )

        if not template_file.exists():
            print(
                f"Warning: B-unit template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Replace class name with placeholder
        class_match = re.search(
            r"class ([A-Za-z0-9]+)BUnits:",
            content,
        )
        if class_match:
            class_name = (
                class_match.group(1)
            )
            content = content.replace(
                f"class {class_name}BUnits:",
                "class {class_name}BUnits:",
            )

        # Save to output file
        output_file = (
            output_path / "b_unit.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for b_unit class."""\n\n'
            )
            f.write(
                f'B_UNIT_TEMPLATE = """{content}"""'
            )

    def _extract_b_unit_creator_and_runner(
        self, output_path: Path
    ) -> None:
        """Extract b_unit_creator_and_runner template."""
        template_file = (
            self.template_path
            / "b_source"
            / "common"
            / "operations"
            / "b_units"
            / "b_unit_creator_and_runner.py"
        )

        if not template_file.exists():
            print(
                f"Warning: B-unit creator and runner template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Save to output file
        output_file = (
            output_path
            / "b_unit_creator_and_runner.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for b_unit creator and runner."""\n\n'
            )
            f.write(
                f'B_UNIT_CREATOR_AND_RUNNER_TEMPLATE = """{content}"""'
            )

    def _extract_conftest(
        self, output_path: Path
    ) -> None:
        """Extract conftest template."""
        template_file = (
            self.template_path
            / "tests"
            / "universal"
            / "e2e"
            / "conftest.py"
        )

        if not template_file.exists():
            print(
                f"Warning: Conftest template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Save to output file
        output_file = (
            output_path / "conftest.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for conftest.py."""\n\n'
            )
            f.write(
                f'CONFTEST_TEMPLATE = """{content}"""'
            )

    def _extract_e2e_test(
        self, output_path: Path
    ) -> None:
        """Extract e2e test template."""
        template_file = (
            self.template_path
            / "tests"
            / "universal"
            / "e2e"
            / "test_aa_b_clearer_pipeline_b_application_runner.py"
        )

        if not template_file.exists():
            print(
                f"Warning: E2E test template file not found: {template_file}"
            )
            return

        with open(
            template_file, "r"
        ) as f:
            content = f.read()

        # Replace domain-specific names with placeholders
        content = content.replace(
            "aa", "{domain_name}"
        )

        # Save to output file
        output_file = (
            output_path / "e2e_test.py"
        )
        with open(
            output_file, "w"
        ) as f:
            f.write(
                f'"""Template for e2e test."""\n\n'
            )
            f.write(
                f'E2E_TEST_TEMPLATE = """{content}"""'
            )

    def _create_init(
        self, output_path: Path
    ) -> None:
        """Create __init__.py with imports for all templates."""
        output_file = (
            output_path / "__init__.py"
        )

        content = """\"\"\"Template constants for pipeline builder.\"\"\"\n
# Import all template modules to make them accessible from the templates package
from .application_runner import APPLICATION_RUNNER_TEMPLATE
from .b_unit import B_UNIT_TEMPLATE
from .b_unit_creator_and_runner import B_UNIT_CREATOR_AND_RUNNER_TEMPLATE
from .conftest import CONFTEST_TEMPLATE
from .e2e_test import E2E_TEST_TEMPLATE
from .pipeline_orchestrator import (
    EMPTY_FUNCTION_BODY as PIPELINE_EMPTY_FUNCTION_BODY,
    ORCHESTRATE_THIN_SLICE_CALL,
    PIPELINE_ORCHESTRATOR_FUNCTION_START,
    PIPELINE_ORCHESTRATOR_IMPORT_TEMPLATE,
)
from .pipeline_runner import PIPELINE_RUNNER_TEMPLATE
from .pipelines_runner import (
    EMPTY_FUNCTION_BODY as PIPELINES_EMPTY_FUNCTION_BODY,
    PIPELINE_IMPORT_TEMPLATE,
    PIPELINE_RUNNER_CALL,
    PIPELINES_RUNNER_FUNCTION_START,
    PIPELINES_RUNNER_TEMPLATE_HEADER,
)
from .stage_orchestrator import (
    B_UNIT_CREATOR_IMPORT as STAGE_B_UNIT_CREATOR_IMPORT,
    B_UNIT_IMPORT_TEMPLATE,
    CREATE_AND_RUN_B_UNIT_CALL as STAGE_CREATE_AND_RUN_B_UNIT_CALL,
    EMPTY_FUNCTION_BODY as STAGE_EMPTY_FUNCTION_BODY,
    ORCHESTRATE_SUB_STAGE_CALL,
    STAGE_ORCHESTRATOR_FUNCTION_START,
    STAGE_ORCHESTRATOR_HEADER,
    SUB_STAGE_IMPORT_TEMPLATE,
)
from .sub_stage_orchestrator import (
    B_UNIT_CREATOR_IMPORT as SUB_STAGE_B_UNIT_CREATOR_IMPORT,
    CREATE_AND_RUN_B_UNIT_CALL as SUB_STAGE_CREATE_AND_RUN_B_UNIT_CALL,
    EMPTY_FUNCTION_BODY as SUB_STAGE_EMPTY_FUNCTION_BODY,
    SUB_STAGE_B_UNIT_IMPORT_TEMPLATE,
    SUB_STAGE_ORCHESTRATOR_FUNCTION_START,
    SUB_STAGE_ORCHESTRATOR_HEADER,
)
from .thin_slice_orchestrator import (
    EMPTY_FUNCTION_BODY as THIN_SLICE_EMPTY_FUNCTION_BODY,
    ORCHESTRATE_STAGE_CALL,
    THIN_SLICE_ORCHESTRATOR_FUNCTION_START,
    THIN_SLICE_ORCHESTRATOR_IMPORT_TEMPLATE,
)"""

        with open(
            output_file, "w"
        ) as f:
            f.write(content)


def update_templates_from_pipeline(
    template_path: str,
    output_directory: str,
) -> None:
    """
    Update template constants from a pipeline template.

    Args:
        template_path: Path to the template pipeline
        output_directory: Path to save the extracted templates
    """
    extractor = TemplateExtractor(
        template_path
    )
    extractor.extract_templates(
        output_directory
    )



"""JSON schema for bclearer pipeline configurations."""

from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
    Union,
)


@dataclass
class BUnitConfig:
    """Configuration for a bUnit in a pipeline stage."""

    name: str


@dataclass
class SubStageConfig:
    """Configuration for a sub-stage in a pipeline stage."""

    name: str
    b_units: List[str]


@dataclass
class StageConfig:
    """Configuration for a stage in a pipeline."""

    name: str
    sub_stages: List[SubStageConfig]
    b_units: List[str]


@dataclass
class ThinSliceConfig:
    """Configuration for a thin slice in a pipeline."""

    name: str
    stages: List[StageConfig]


@dataclass
class PipelineConfig:
    """Configuration for a pipeline in a domain."""

    name: str
    thin_slices: List[ThinSliceConfig]


@dataclass
class DomainConfig:
    """Top-level configuration for a domain with pipelines."""

    domain_name: str
    pipelines: List[PipelineConfig]


def validate_pipeline_config(
    config_dict: Dict,
) -> DomainConfig:
    """
    Validates the pipeline configuration dictionary and converts it to a DomainConfig object.

    Args:
        config_dict: Dictionary containing the pipeline configuration

    Returns:
        DomainConfig object
    """
    try:
        domain_name = config_dict.get(
            "domain_name"
        )
        if not domain_name:
            raise ValueError(
                "domain_name is required"
            )

        pipelines_data = (
            config_dict.get(
                "pipelines", []
            )
        )
        pipelines = []

        for (
            pipeline_data
        ) in pipelines_data:
            pipeline_name = (
                pipeline_data.get(
                    "name"
                )
            )
            if not pipeline_name:
                raise ValueError(
                    "Pipeline name is required"
                )

            thin_slices_data = (
                pipeline_data.get(
                    "thin_slices", []
                )
            )
            thin_slices = []

            for (
                thin_slice_data
            ) in thin_slices_data:
                thin_slice_name = (
                    thin_slice_data.get(
                        "name"
                    )
                )
                if not thin_slice_name:
                    raise ValueError(
                        f"Thin slice name is required in pipeline '{pipeline_name}'"
                    )

                stages_data = (
                    thin_slice_data.get(
                        "stages", []
                    )
                )
                stages = []

                for (
                    stage_data
                ) in stages_data:
                    stage_name = (
                        stage_data.get(
                            "name"
                        )
                    )
                    if not stage_name:
                        raise ValueError(
                            f"Stage name is required in thin slice '{thin_slice_name}'"
                        )

                    sub_stages_data = stage_data.get(
                        "sub_stages", []
                    )
                    sub_stages = []

                    for (
                        sub_stage_data
                    ) in (
                        sub_stages_data
                    ):
                        sub_stage_name = sub_stage_data.get(
                            "name"
                        )
                        if (
                            not sub_stage_name
                        ):
                            raise ValueError(
                                f"Sub-stage name is required in stage '{stage_name}'"
                            )

                        b_units = sub_stage_data.get(
                            "b_units",
                            [],
                        )

                        sub_stages.append(
                            SubStageConfig(
                                name=sub_stage_name,
                                b_units=b_units,
                            )
                        )

                    b_units = (
                        stage_data.get(
                            "b_units",
                            [],
                        )
                    )

                    stages.append(
                        StageConfig(
                            name=stage_name,
                            sub_stages=sub_stages,
                            b_units=b_units,
                        )
                    )

                thin_slices.append(
                    ThinSliceConfig(
                        name=thin_slice_name,
                        stages=stages,
                    )
                )

            pipelines.append(
                PipelineConfig(
                    name=pipeline_name,
                    thin_slices=thin_slices,
                )
            )

        return DomainConfig(
            domain_name=domain_name,
            pipelines=pipelines,
        )
    except Exception as e:
        raise ValueError(
            f"Invalid pipeline configuration: {str(e)}"
        )


def get_sample_config() -> Dict:
    """
    Returns a sample pipeline configuration.

    Returns:
        Dict containing a sample pipeline configuration
    """
    return {
        "domain_name": "example_domain",
        "pipelines": [
            {
                "name": "pipeline_name",
                "thin_slices": [
                    {
                        "name": "thin_slice_1",
                        "stages": [
                            {
                                "name": "1c_collect",
                                "sub_stages": [],
                                "b_units": [
                                    "ca_b_unit",
                                    "cb_b_unit",
                                ],
                            },
                            {
                                "name": "2l_load",
                                "sub_stages": [],
                                "b_units": [
                                    "la_b_unit",
                                    "lb_b_unit",
                                ],
                            },
                            {
                                "name": "3e_evolve",
                                "sub_stages": [
                                    {
                                        "name": "sub_stage_1",
                                        "b_units": [
                                            "ea1_b_unit",
                                            "ea2_b_unit",
                                        ],
                                    },
                                    {
                                        "name": "sub_stage_2",
                                        "b_units": [
                                            "eb1_b_unit",
                                            "eb2_b_unit",
                                        ],
                                    },
                                ],
                                "b_units": [],
                            },
                            {
                                "name": "4a_assimilate",
                                "sub_stages": [],
                                "b_units": [
                                    "aa_b_unit",
                                    "ab_b_unit",
                                ],
                            },
                            {
                                "name": "5r_reuse",
                                "sub_stages": [],
                                "b_units": [
                                    "ra_b_unit",
                                    "rb_b_unit",
                                ],
                            },
                        ],
                    }
                ],
            }
        ],
    }



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)

LIST_OF_NF_UUID_COLUMN_NAMES = [
    NfColumnTypes.NF_UUIDS.column_name,
    NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
    NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name,
    NfEaComColumnTypes.ELEMENTS_CLASSIFIER.column_name,
    NfEaComColumnTypes.CLASSIFIERS_CONTAINING_EA_ELEMENT.column_name,
    NfEaComColumnTypes.PACKAGEABLE_OBJECTS_PARENT_EA_ELEMENT.column_name,
    # It is used in the data both as uuid field and uuid list field
    NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name,
    # It is used in the data both as uuid field and uuid list field
    NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name,
    "provider",
    "dependent",
    "specialisation",
    "generalisation",
    "child",
    "parent",
    "ea_client",
    "ea_stereotype",
    NfEaComColumnTypes.STEREOTYPE_EA_STEREOTYPE_GROUP.column_name,
    NfEaComColumnTypes.STEREOTYPE_CLIENT_NF_UUIDS.column_name,
    "stereotype_nf_uuids",
]

LIST_OF_NF_UUID_LISTS_COLUMN_NAMES = [
    NfEaComColumnTypes.CLASSIFIERS_ALL_COMPONENT_EA_ATTRIBUTES.column_name,
    # It is used in the data both as uuid field and uuid list field
    NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name,
    # It is used in the data both as uuid field and uuid list field
    NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS,
    NfEaComColumnTypes.CLASSIFIERS_ALL_COMPONENT_EA_OPERATIONS.column_name,
    NfEaComColumnTypes.ELEMENTS_CONTAINED_EA_DIAGRAMS.column_name,
    NfEaComColumnTypes.ELEMENTS_CONTAINED_EA_CLASSIFIERS.column_name,
    NfEaComColumnTypes.STEREOTYPEABLE_OBJECTS_EA_OBJECT_STEREOTYPES.column_name,
    NfEaComColumnTypes.PACKAGES_CONTAINED_EA_PACKAGES.column_name,
    "paths",
]
